import datetime as dt
import json

from textwrap import dedent

import nbformat

from panel import Column
from panel.chat import ChatStep

from lumen.ai.views import LumenOutput
from lumen.config import config
from lumen.pipeline import Pipeline
from lumen.views import View


def make_preamble(preamble):
    now = dt.datetime.now()
    header = nbformat.v4.new_markdown_cell(source=f'# Lumen.ai - Chat Logs {now}')
    source = (preamble + dedent(
        """
        import lumen as lm
        import panel as pn

        pn.extension('tabulator')
        """
    )).strip()
    imports = nbformat.v4.new_code_cell(source=source)
    return [header, imports]

def format_markdown(msg):
    if msg.avatar.startswith('https://'):
        avatar = f'<img src="{msg.avatar}" width=45 height=45></img>'
    else:
        avatar = f'<span>{msg.avatar}</span>'
    header = f'<div style="display: flex; flex-direction: row; font-weight: bold; font-size: 2em;">{avatar}<span style="margin-left: 0.5em">{msg.user}</span></div>'
    prefix = '\n' if msg.user == 'User' else '\n> '
    content = prefix.join(msg.serialize().split('\n'))
    return [nbformat.v4.new_markdown_cell(source=f'{header}\n{prefix}{content}')]

def format_output(msg):
    output = msg.object
    code = []
    with config.param.update(serializer='csv'):
        spec = json.dumps(output.component.to_spec(), indent=2).replace('true', 'True').replace('false', 'False')
    if isinstance(output.component, Pipeline):
        code.extend([
            f'pipeline = lm.Pipeline.from_spec({spec})',
            'pipeline'
        ])
    elif isinstance(output.component, View):
        code.extend([
            f'view = lm.View.from_spec({spec})',
            'view'
        ])
    return [nbformat.v4.new_code_cell(source='\n'.join(code))]

def export_notebook(assistant, preamble: str = ""):
    cells = make_preamble(preamble)
    for msg in assistant.interface.objects:
        if msg.user == 'Help':
            continue
        elif isinstance(msg.object, str):
            cells += format_markdown(msg)
        elif isinstance(msg.object, LumenOutput):
            cells += format_output(msg)
        elif isinstance(msg.object, Column):
            for obj in msg.object:
                if isinstance(obj, ChatStep):
                    continue
                cells += format_output(obj)

    nb = nbformat.v4.new_notebook(cells=cells)
    return nbformat.v4.writes(nb)



from pathlib import Path

from .config import DEFAULT_EMBEDDINGS_PATH


class Embeddings:

    def add_directory(self, data_dir: Path):
        raise NotImplementedError

    def query(self, query_texts: str) -> list:
        raise NotImplementedError


class ChromaDb(Embeddings):

    def __init__(self, collection: str, persist_dir: str = DEFAULT_EMBEDDINGS_PATH):
        import chromadb
        self.client = chromadb.PersistentClient(path=str(persist_dir / collection))
        self.collection = self.client.get_or_create_collection(collection)

    def add_directory(self, data_dir: Path, file_type='json'):
        add_kwargs = {
            "ids": [],
            "documents": [],
        }
        for i, path in enumerate(data_dir.glob(f"**/*.{file_type}")):
            add_kwargs["ids"].append(f"{i}")
            add_kwargs["documents"].append(path.read_text())
        self.collection.add(**add_kwargs)

    def query(self, query_texts: str) -> list:
        return self.collection.query(query_texts=query_texts)["documents"]



import io

import pandas as pd
import panel as pn
import param

from lumen.ai.memory import memory
from lumen.sources.duckdb import DuckDBSource


class TableControls(pn.viewable.Viewer):

    filename = param.String(default="", doc="Filename")
    table = param.String(default="", doc="Table name")
    extension = param.String(default="", doc="File extension")
    sheet = param.ObjectSelector(default=None, objects=[], doc="Sheet")

    _load = param.Event(doc="Load table")

    def __init__(self, file: io.BytesIO, **params):
        filename = params["filename"]
        if "/" in filename:
            filename, extension = filename.rsplit("/")[-1].split(".", maxsplit=1)
        elif not self.filename and filename:
            filename, extension = filename.rsplit(".", maxsplit=1)
        params["filename"] = filename
        params["table"] = filename.replace("-", "_")
        params["extension"] = extension
        super().__init__(**params)
        self.file = file
        self._name_input = pn.widgets.TextInput.from_param(
            self.param.table, name="Table name"
        )
        self._sheet_select = pn.widgets.Select.from_param(
            self.param.sheet, name="Sheet", visible=False
        )
        self.box = pn.FlexBox(
            self._name_input,
            self._sheet_select,
        )
        self.param.trigger("table")  # fix name
        self.param.trigger("_load")  # to prevent blocking

    @param.depends("_load", watch=True)
    async def _select_sheet(self):
        if not self.extension.endswith("xlsx"):
            return
        import openpyxl

        wb = openpyxl.load_workbook(self.file, read_only=True)
        with param.parameterized.batch_call_watchers(self):
            self.param.sheet.objects = wb.sheetnames
            self.sheet = wb.sheetnames[0]
            self._sheet_select.visible = True

    @param.depends("table", watch=True)
    async def _replace_with_underscore(self):
        self.table = "".join(
            c if c.isalnum() or c == "." else "_" for c in self.table
        ).strip("_").lower()

    def __panel__(self):
        return self.box


class SourceControls(pn.viewable.Viewer):

    add = param.Event(doc="Add tables")
    multiple = param.Boolean(default=False, doc="Allow multiple files")
    replace_controls = param.Boolean(default=False, doc="Replace controls")
    select_existing = param.Boolean(default=True, doc="Select existing table")

    _last_table = param.String(default="", doc="Last table added")

    def __init__(self, **params):
        super().__init__(**params)

        self.tables_tabs = pn.Tabs(sizing_mode="stretch_width")
        self._file_input = pn.widgets.FileDropper(
            height=90,
            multiple=self.param.multiple,
            margin=0,
            sizing_mode="stretch_width",
            # accepted_filetypes=[".csv", ".parquet", ".parq", ".json", ".xlsx"],
        )
        self._file_input.param.watch(self._generate_table_controls, "value")
        self._upload_tabs = pn.Tabs(sizing_mode="stretch_width")

        self._input_tabs = pn.Tabs(
            ("Upload", pn.Column(self._file_input, self._upload_tabs)),
            sizing_mode="stretch_both",
        )

        if self.select_existing:
            nested_sources_tables = {
                source.name: source.get_tables() for source in memory["available_sources"]
            }
            first_table = {k: nested_sources_tables[k][0] for k in list(nested_sources_tables)[:1]}
            self._select_table = pn.widgets.NestedSelect(
                name="Table",
                value=first_table,
                options=nested_sources_tables,
                levels=["source", "table"],
                sizing_mode="stretch_width",
            )
            self._input_tabs.append(("Select", self._select_table))
            self._select_table.param.watch(self._generate_table_controls, "value")

        self._add_button = pn.widgets.Button.from_param(
            self.param.add,
            name="Use tables",
            icon="table-plus",
            visible=False,
            button_type="success",
        )
        self.menu = pn.Column(
            self._input_tabs,
            self._add_button,
            self.tables_tabs,
            sizing_mode="stretch_width",
        )

        self._table_controls = []

    def _generate_table_controls(self, event):
        if self._input_tabs.active == 0:
            self._upload_tabs.clear()
            self._table_controls.clear()
            for filename, file in self._file_input.value.items():
                table_controls = TableControls(
                    io.BytesIO(file) if isinstance(file, bytes) else io.StringIO(file),
                    filename=filename,
                )
                self._upload_tabs.append((filename, table_controls))
                self._table_controls.append(table_controls)

            if len(self._upload_tabs) > 0:
                self._add_button.visible = True
            else:
                self._add_button.visible = False
        elif self._input_tabs.active == 1:
            self._add_button.visible = True

    def _add_table(
        self,
        duckdb_source: DuckDBSource,
        file: io.BytesIO | io.StringIO,
        table_controls: TableControls,
    ):
        extension = table_controls.extension
        table = table_controls.table
        if extension.endswith("csv"):
            df = pd.read_csv(file, parse_dates=True)
        elif extension.endswith(("parq", "parquet")):
            df = pd.read_parquet(file)
        elif extension.endswith("json"):
            df = pd.read_json(file)
        elif extension.endswith("xlsx"):
            sheet = table_controls.sheet
            df = pd.read_excel(file, sheet_name=sheet)
        else:
            raise ValueError(f"Unsupported file extension: {extension}")

        duckdb_source._connection.from_df(df).to_view(table)
        duckdb_source.tables[table] = f"SELECT * FROM {table}"
        memory["current_source"] = duckdb_source
        memory["current_table"] = table
        memory["available_sources"].add(duckdb_source)
        self._last_table = table

    @param.depends("add", watch=True)
    def add_tables(self):
        with self.menu.param.update(loading=True):
            duckdb_source = DuckDBSource(uri=":memory:", ephemeral=True)
            if duckdb_source.tables is None:
                duckdb_source.tables = {}
            if self._input_tabs.active == 0:
                for i in range(len(self._upload_tabs)):
                    table_controls = self._table_controls[i]
                    self._add_table(duckdb_source, table_controls.file, table_controls)

                if self.replace_controls:
                    src = memory["current_source"]
                    self.tables_tabs[:] = [
                        (
                            t,
                            pn.widgets.Tabulator(
                                src.get(t), sizing_mode="stretch_both"
                            ),
                        )
                        for t in src.get_tables()
                    ]
                    self.menu[0].visible = False
                    self._add_button.visible = False
            elif self.select_existing and self._input_tabs.active == 1:
                table = self._select_table.value["table"]
                duckdb_source.tables[table] = f"SELECT * FROM {table}"
                memory["current_source"] = duckdb_source
                memory["current_table"] = table
                memory["available_sources"].add(duckdb_source)
                self._last_table = table

    def __panel__(self):
        return self.menu



from __future__ import annotations

import re

from typing import TYPE_CHECKING, Any

import duckdb
import param

from ..config import config
from ..serializers import Serializer
from ..transforms import Filter
from ..transforms.sql import (
    SQLCast, SQLDistinct, SQLFilter, SQLLimit, SQLMinMax,
)
from .base import (
    BaseSQLSource, Source, cached, cached_schema,
)

if TYPE_CHECKING:
    import pandas as pd


class DuckDBSource(BaseSQLSource):
    """
    DuckDBSource provides a simple wrapper around the DuckDB SQL
    connector.

    To specify tables to be queried provide a list or dictionary of
    tables. A SQL expression to fetch the data from the table will
    then be generated using the `sql_expr`, e.g. if we specify a local
    table flights.db the default sql_expr `SELECT * FROM {table}` will
    expand that to `SELECT * FROM flights.db`. If you want to specify
    a full SQL expression as a table you must change the `sql_expr` to
    '{table}' ensuring no further templating is applied.

    Note that certain functionality in DuckDB requires modules to be
    loaded before making a query. These can be specified using the
    `initializers` parameter, providing the ability to define DuckDb
    statements to be run when initializing the connection.
    """

    filter_in_sql = param.Boolean(default=True, doc="""
        Whether to apply filters in SQL or in-memory.""")

    initializers = param.List(default=[], doc="""
        SQL statements to run to initialize the connection.""")

    mirrors = param.Dict(default={}, doc="""
        Mirrors the tables into the DuckDB database. The mirrors
        should define a mapping from the table names to the source of
        the mirror which may be defined as a Pipeline or a tuple of
        the source and the table.""")

    load_schema = param.Boolean(default=True, doc="Whether to load the schema")

    sql_expr = param.String(default='SELECT * FROM {table}', doc="""
        The SQL expression to execute.""")

    ephemeral = param.Boolean(default=False, doc="""
        Whether the data is ephemeral, i.e. manually inserted into the
        DuckDB table or derived from real data.""")

    tables = param.ClassSelector(class_=(list, dict), doc="""
        List or dictionary of tables.""")

    uri = param.String(doc="The URI of the DuckDB database")

    source_type = 'duckdb'

    dialect = 'duckdb'

    def __init__(self, **params):
        connection = params.pop('_connection', None)
        super().__init__(**params)
        if connection:
            self._connection = connection
        else:
            self._connection = duckdb.connect(self.uri)
            for init in self.initializers:
                self._connection.execute(init)
        for table, mirror in self.mirrors.items():
            if isinstance(mirror, tuple):
                source, src_table = mirror
                df = source.get(src_table)
            else:
                df = mirror.data
                def update(e, table=table):
                    self._connection.from_df(e.new).to_view(table)
                    self._set_cache(e.new, table)
                mirror.param.watch(update, 'data')
            try:
                self._connection.from_df(df).to_view(table)
            except (duckdb.CatalogException, duckdb.ParserException):
                continue

    @property
    def connection(self):
        return self._connection

    @classmethod
    def _recursive_resolve(
        cls, spec: dict[str, Any], source_type: type[Source]
    ) -> tuple[dict[str, Any], dict[str, Any]]:
        mirrors = spec.pop('mirrors', {})
        spec, refs = super()._recursive_resolve(spec, source_type)
        resolved_mirrors = {}
        for table, (src_spec, src_table) in mirrors.items():
            source = cls.from_spec(src_spec)
            resolved_mirrors[table] = (source, src_table)
        spec['mirrors'] = resolved_mirrors
        return spec, refs

    def _serialize_tables(self):
        tables = {}
        for t in self.get_tables():
            tdf = self.get(t)
            serializer = Serializer._get_type(config.serializer)()
            tables[t] = serializer.serialize(tdf)
        return tables

    def to_spec(self, context: dict[str, Any] | None = None) -> dict[str, Any]:
        spec = super().to_spec(context)
        if self.ephemeral:
            spec['tables'] = self._serialize_tables()
        if 'mirrors' not in spec:
            return spec
        mirrors = {}
        for table, (source, src_table) in spec['mirrors'].items():
            src_spec = source.to_spec(context=context)
            mirrors[table] = (src_spec, src_table)
        spec['mirrors'] = mirrors
        return spec

    @classmethod
    def from_df(cls, tables: dict[str, pd.DataFrame], **kwargs):
        """
        Creates an ephemeral, in-memory DuckDBSource containing the
        supplied dataframe.

        Arguments
        ---------
        tables: dict[str, pandas.DataFrame]
            A dictionary mapping from table names to DataFrames
        kwargs: any
            Additional keyword arguments for the source

        Returns
        -------
        source: DuckDBSource
        """
        source = DuckDBSource(uri=':memory:', ephemeral=True, **kwargs)
        table_defs = {}
        for name, df in tables.items():
            source._connection.from_df(df).to_view(name)
            table_defs[name] = source.sql_expr.format(table=name)
        source.tables = table_defs
        return source

    @classmethod
    def from_spec(cls, spec: dict[str, Any] | str) -> Source:
        if spec.get('ephemeral') and 'tables' in spec:
            ephemeral_tables = spec['tables']
            spec['tables'] = {}
        else:
            ephemeral_tables = {}
        source = super().from_spec(spec)
        if not ephemeral_tables:
            return source
        new_tables = {}
        for t, data in ephemeral_tables.items():
            df = Serializer.deserialize(data)
            try:
                source._connection.from_df(df).to_view(t)
            except duckdb.ParserException:
                continue
            new_tables[t] = source.sql_expr.format(table=t)
        source.tables = new_tables
        return source

    def create_sql_expr_source(
        self, tables: dict[str, str], materialize: bool = True, **kwargs
    ):
        """
        Creates a new SQL Source given a set of table names and
        corresponding SQL expressions.

        Arguments
        ---------
        tables: dict[str, str]
            Mapping from table name to SQL expression.
        materialize: bool
            Whether to materialize new tables
        kwargs: any
            Additional keyword arguments.

        Returns
        -------
        source: DuckDBSource
        """
        params = dict(self.param.values(), **kwargs)
        params['tables'] = tables
        # Reuse connection unless it has changed
        if 'uri' not in kwargs and 'initializers' not in kwargs:
            params['_connection'] = self._connection
        source = type(self)(**params)
        if not materialize:
            return source
        existing = self.get_tables()
        for table, sql_expr in tables.copy().items():
            if table in existing or sql_expr == self.sql_expr.format(table=f'"{table}"'):
                continue
            table_expr = f'CREATE TEMP TABLE "{table}" AS ({sql_expr})'
            try:
                self._connection.execute(table_expr)
            except duckdb.CatalogException as e:
                pattern = r"Table with name\s(\S+)"
                match = re.search(pattern, str(e))
                if match and isinstance(self.tables, dict):
                    name = match.group(1)
                    real = self.tables[name]
                    if 'select' not in real.lower() and not real.startswith('"'):
                        real = f'"{real}"'
                    else:
                        real = f'({real})'
                    table_expr = table_expr.replace(f'FROM {name}', f'FROM {real}')
                    source.tables[table] = sql_expr.replace(f'FROM {name}', f'FROM {real}')
                    self._connection.execute(table_expr)
                else:
                    raise e
        return source

    def get_tables(self):
        if isinstance(self.tables, (dict, list)):
            return list(self.tables)
        return [t[0] for t in self._connection.execute('SHOW TABLES').fetchall()]

    def get_sql_expr(self, table: str):
        if isinstance(self.tables, dict):
            table = self.tables[table]
        if '(' not in table and ')' not in table:
            table = f'"{table}"'
        if 'select ' in table.lower():
            sql_expr = table
        else:
            sql_expr = self.sql_expr.format(table=table)
        return sql_expr.rstrip(";")

    @cached
    def get(self, table, **query):
        query.pop('__dask', None)
        sql_expr = self.get_sql_expr(table)
        sql_transforms = query.pop('sql_transforms', [])
        conditions = list(query.items())
        if self.filter_in_sql:
            sql_transforms = [SQLFilter(conditions=conditions)] + sql_transforms
        for st in sql_transforms:
            sql_expr = st.apply(sql_expr)
        df = self._connection.execute(sql_expr).fetch_df(date_as_object=True)
        if not self.filter_in_sql:
            df = Filter.apply_to(df, conditions=conditions)
        return df

    @cached_schema
    def get_schema(
        self, table: str | None = None, limit: int | None = None
    ) -> dict[str, dict[str, Any]] | dict[str, Any]:
        if table is None:
            tables = self.get_tables()
        else:
            tables = [table]

        schemas = {}
        sql_limit = SQLLimit(limit=limit or 3)
        for entry in tables:
            if not self.load_schema:
                schemas[entry] = {}
                continue
            sql_expr = self.get_sql_expr(entry)
            data = self._connection.execute(sql_limit.apply(sql_expr)).fetch_df()
            schema, casts = self._compute_subset_schema(data)
            schemas[entry] = schema
            if limit:
                continue

            enums, min_maxes = [], []
            for name, col_schema in schema.items():
                if 'inclusiveMinimum' in col_schema or name in casts:
                    min_maxes.append(name)
                elif 'enum' in col_schema:
                    enums.append(name)
            for col in enums:
                distinct_expr = SQLDistinct(columns=[col]).apply(sql_expr)
                distinct_expr = ' '.join(distinct_expr.splitlines())
                distinct = self._connection.execute(distinct_expr).fetch_df()
                schema[col]['enum'] = distinct[col].tolist()

            if not min_maxes:
                continue

            if casts:
                sql_expr = SQLCast(columns=casts, force=True).apply(sql_expr)
            minmax_expr = SQLMinMax(columns=min_maxes).apply(sql_expr)
            minmax_expr = ' '.join(minmax_expr.splitlines())
            minmax_data = self._connection.execute(minmax_expr).fetch_df()
            for col in min_maxes:
                kind = data[col].dtype.kind
                if kind in 'iu':
                    cast = int
                elif kind == 'f':
                    cast = float
                elif kind == 'M':
                    cast = str
                else:
                    cast = lambda v: v
                schema[col]['inclusiveMinimum'] = cast(minmax_data[f'{col}_min'].iloc[0])
                schema[col]['inclusiveMaximum'] = cast(minmax_data[f'{col}_max'].iloc[0])
        return schemas if table is None else schemas[table]



from __future__ import annotations

import sqlite3

import param


class ChatLogs(param.Parameterized):

    filename = param.String(default="chat_logs.db")

    def __init__(self, **params):
        super().__init__(**params)
        self.conn = sqlite3.connect(self.filename)
        self.cursor = self.conn.cursor()
        self.cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS logs (
                session_id TEXT,
                message_id TEXT PRIMARY KEY,
                message_index INTEGER,
                message_user TEXT,
                message_content TEXT,
                liked BOOLEAN DEFAULT FALSE,
                disliked BOOLEAN DEFAULT FALSE,
                removed BOOLEAN DEFAULT FALSE,
                timestamp TEXT DEFAULT CURRENT_TIMESTAMP
            )
            """
        )
        self.conn.commit()

    def upsert(
        self,
        session_id,
        message_id,
        message_index,
        message_user,
        message_content,
    ):
        UPSERT_SCHEMA = """
        INSERT INTO logs (session_id, message_id, message_index, message_user, message_content)
        VALUES (?, ?, ?, ?, ?)
        ON CONFLICT (message_id)
        DO UPDATE SET
        session_id = excluded.session_id,
        message_id = excluded.message_id,
        message_index = excluded.message_index,
        message_user = excluded.message_user,
        message_content = excluded.message_content
        """
        try:
            self.cursor.execute(
                UPSERT_SCHEMA,
                (
                    session_id,
                    message_id,
                    message_index,
                    message_user,
                    message_content,
                ),
            )
        except Exception:
            try:
                self.cursor.execute(
                    UPSERT_SCHEMA,
                    (
                        session_id,
                        message_id,
                        message_index,
                        message_user,
                        str(message_content),
                    ),
                )
            except Exception:
                print("Failed to insert message")
                return
            self.conn.commit()

    def update_status(self, message_id, liked=None, disliked=None, removed=None):
        self.cursor.execute(
            """
            UPDATE logs
            SET liked = COALESCE(?, liked), disliked = COALESCE(?, disliked), removed = COALESCE(?, removed)
            WHERE message_id = ?
            """,
            (liked, disliked, removed, message_id),
        )
        self.conn.commit()



import asyncio
import inspect
import time

from functools import wraps
from pathlib import Path

import jinja2
import pandas as pd

from lumen.pipeline import Pipeline
from lumen.sources.base import Source

from .config import THIS_DIR, UNRECOVERABLE_ERRORS


def render_template(template, **context):
    template_path = Path(template)
    if not template_path.exists():
        template_path = THIS_DIR / "prompts" / template
    template_contents = template_path.read_text()
    template = jinja2.Template(template_contents, undefined=jinja2.StrictUndefined)
    # raise if missing keys in context for template
    return template.render(context).strip()


def retry_llm_output(retries=3, sleep=1):
    """
    Retry a function that returns a response from the LLM API.
    If the function raises an exception, retry it up to `retries` times.
    If `sleep` is provided, wait that many seconds between retries.
    If an error occurs, pass the error message into kwargs for the function
    to manually handle by including it.
    """

    def decorator(func):
        if inspect.iscoroutinefunction(func):

            @wraps(func)
            async def async_wrapper(*args, **kwargs):
                errors = []
                for i in range(retries):
                    if errors:
                        kwargs["errors"] = errors

                    try:
                        output = await func(*args, **kwargs)
                        if not output:
                            raise Exception("No valid output from LLM.")
                        return output
                    except Exception as e:
                        if isinstance(e, UNRECOVERABLE_ERRORS) or i == retries - 1:
                            raise
                        errors.append(str(e))
                        if sleep:
                            await asyncio.sleep(sleep)

            return async_wrapper
        else:

            @wraps(func)
            def sync_wrapper(*args, **kwargs):
                errors = []
                for i in range(retries):
                    if errors:
                        kwargs["errors"] = errors
                    try:
                        output = func(*args, **kwargs)
                        if not output:
                            raise Exception("No valid output from LLM.")
                        return output
                    except Exception as e:
                        if isinstance(e, UNRECOVERABLE_ERRORS) or i == retries - 1:
                            raise
                        errors.append(str(e))
                        if sleep:
                            time.sleep(sleep)

            return sync_wrapper

    return decorator


def format_schema(schema):
    formatted = {}
    for field, spec in schema.items():
        if "enum" in spec and len(spec["enum"]) > 5:
            spec["enum"] = spec["enum"][:5] + ["..."]
        if "type" in spec:
            if spec["type"] == "string":
                spec["type"] = "str"
            elif spec["type"] == "integer":
                spec["type"] = "int"
            elif spec["type"] == "boolean":
                spec["type"] = "bool"
        formatted[field] = spec
    return formatted


def get_schema(
    source: Source | Pipeline,
    table: str | None = None,
    include_min_max: bool = True,
    include_enum: bool = True,
    **get_kwargs
):
    if isinstance(source, Pipeline):
        schema = source.get_schema()
    else:
        if "limit" not in get_kwargs:
            get_kwargs["limit"] = 100
        schema = source.get_schema(table, **get_kwargs)
    schema = dict(schema)

    if include_min_max:
        for field, spec in schema.items():
            if "inclusiveMinimum" in spec:
                spec["min"] = spec.pop("inclusiveMinimum")
            if "inclusiveMaximum" in spec:
                spec["max"] = spec.pop("inclusiveMaximum")
    else:
        for field, spec in schema.items():
            if "inclusiveMinimum" in spec:
                spec.pop("inclusiveMinimum")
            if "inclusiveMaximum" in spec:
                spec.pop("inclusiveMaximum")
            if "min" in spec:
                spec.pop("min")
            if "max" in spec:
                spec.pop("max")

    if not include_enum:
        for field, spec in schema.items():
            if "enum" in spec:
                spec.pop("enum")

    schema = format_schema(schema)
    return schema


def describe_data(df: pd.DataFrame) -> str:
    def format_float(num):
        if pd.isna(num):
            return num
        # if is integer, round to 0 decimals
        if num == int(num):
            return f"{int(num)}"
        elif 0.01 <= abs(num) < 100:
            return f"{num:.1f}"  # Regular floating-point notation with two decimals
        else:
            return f"{num:.1e}"  # Exponential notation with two decimals

    size = df.size
    shape = df.shape
    if size < 250:
        return df

    is_summarized = False
    if shape[0] > 5000:
        is_summarized = True
        df = df.sample(5000)

    df = df.sort_index()

    for col in df.columns:
        if isinstance(df[col].iloc[0], pd.Timestamp):
            df[col] = pd.to_datetime(df[col])

    describe_df = df.describe(percentiles=[])
    columns_to_drop = ["min", "max"] # present if any numeric
    columns_to_drop = [col for col in columns_to_drop if col in describe_df.columns]
    df_describe_dict = describe_df.drop(columns=columns_to_drop).to_dict()

    for col in df.select_dtypes(include=["object"]).columns:
        if col not in df_describe_dict:
            df_describe_dict[col] = {}
        df_describe_dict[col]["nunique"] = df[col].nunique()
        try:
            df_describe_dict[col]["lengths"] = {
                "max": df[col].str.len().max(),
                "min": df[col].str.len().min(),
                "mean": float(df[col].str.len().mean()),
            }
        except AttributeError:
            pass

    for col in df.columns:
        if col not in df_describe_dict:
            df_describe_dict[col] = {}
        df_describe_dict[col]["nulls"] = int(df[col].isnull().sum())

    # select datetime64 columns
    for col in df.select_dtypes(include=["datetime64"]).columns:
        for key in df_describe_dict[col]:
            df_describe_dict[col][key] = str(df_describe_dict[col][key])
        df[col] = df[col].astype(str)  # shorten output

    # select all numeric columns and round
    for col in df.select_dtypes(include=["int64", "float64"]).columns:
        for key in df_describe_dict[col]:
            df_describe_dict[col][key] = format_float(df_describe_dict[col][key])

    for col in df.select_dtypes(include=["float64"]).columns:
        df[col] = df[col].apply(format_float)

    data = {
        "summary": {
            "total_table_cells": size,
            "total_shape": shape,
            "is_summarized": is_summarized,
        },
        "stats": df_describe_dict,
    }
    return data


def clean_sql(sql_expr):
    """
    Cleans up a SQL expression generated by an LLM by removing
    backticks, fencing and extraneous space and semi-colons.
    """
    return sql_expr.replace("```sql", "").replace("```", "").replace('`', '"').strip().rstrip(";")



import pytest

try:
    import intake_dremio
except Exception:
    pytest.skip('intake-dremio not available.', allow_module_level=True)

from lumen.sources.intake_dremio import IntakeDremioSQLSource


def test_intake_dremio_resolve_module_type():
    assert IntakeDremioSQLSource._get_type('lumen.sources.intake_dremio.IntakeDremioSQLSource') is IntakeDremioSQLSource
    assert IntakeDremioSQLSource.source_type == 'intake_dremio_sql'


def test_intake_dremio_init():
    credentials = dict(
        cert="./test.ipynb",
        uri="uri",
        tls=True,
        username="username",
        password="password",
    )
    source = IntakeDremioSQLSource(
        tables={"table": "SELECT * FROM table", "table2": "SELECT * FROM table2 LIMIT 3;"},
        **credentials
    )
    assert source.tables == {"table": "SELECT * FROM table", "table2": "SELECT * FROM table2 LIMIT 3;"}
    assert source.cat == {
        "table": intake_dremio.DremioSource(
            sql_expr="SELECT * FROM table",
            **credentials
        ),
        "table2": intake_dremio.DremioSource(
            sql_expr="SELECT * FROM table2 LIMIT 3;",
            **credentials
        )
    }



from __future__ import annotations

import asyncio

from io import StringIO
from typing import Literal

import param

from panel import bind
from panel.chat import ChatInterface, ChatStep
from panel.layout import Column, FlexBox, Tabs
from panel.pane import HTML, Markdown
from panel.viewable import Viewer
from panel.widgets import Button, FileDownload
from pydantic import create_model
from pydantic.fields import FieldInfo

from .agents import (
    Agent, AnalysisAgent, ChatAgent, SQLAgent,
)
from .config import DEMO_MESSAGES, GETTING_STARTED_SUGGESTIONS
from .export import export_notebook
from .llm import Llama, Llm
from .logs import ChatLogs
from .memory import memory
from .models import Validity
from .utils import get_schema, render_template, retry_llm_output


class Assistant(Viewer):
    """
    An Assistant handles multiple agents.
    """

    agents = param.List(default=[ChatAgent])

    llm = param.ClassSelector(class_=Llm, default=Llama())

    interface = param.ClassSelector(class_=ChatInterface)

    suggestions = param.List(default=GETTING_STARTED_SUGGESTIONS)

    demo_inputs = param.List(default=DEMO_MESSAGES)

    notebook_preamble = param.String()

    sidebar_widgets = param.List()

    logs_filename = param.String()

    def __init__(
        self,
        llm: Llm | None = None,
        interface: ChatInterface | None = None,
        agents: list[Agent | type[Agent]] | None = None,
        logs_filename: str = "",
        **params,
    ):

        def on_message(message, instance):
            def update_on_reaction(reactions):
                self._logs.update_status(
                    message_id=message_id,
                    liked="like" in reactions,
                    disliked="dislike" in reactions,
                )

            bind(update_on_reaction, message.param.reactions, watch=True)
            message_id = id(message)
            message_index = len(instance) - 1
            self._logs.upsert(
                session_id=self._session_id,
                message_id=message_id,
                message_index=message_index,
                message_user=message.user,
                message_content=message.object,
            )

        def on_undo(instance, _):
            count = instance._get_last_user_entry_index()
            messages = instance[-count:]
            for message in messages:
                self._logs.update_status(message_id=id(message), removed=True)

        def on_rerun(instance, _):
            count = instance._get_last_user_entry_index() - 1
            messages = instance[-count:]
            for message in messages:
                self._logs.update_status(message_id=id(message), removed=True)

        def download_notebook():
            nb = export_notebook(self, preamble=self.notebook_preamble)
            return StringIO(nb)

        if interface is None:
            interface = ChatInterface(
                callback=self._chat_invoke, load_buffer=5,
            )
        else:
            interface.callback = self._chat_invoke
        interface.callback_exception = "verbose"
        interface.message_params["reaction_icons"] = {"like": "thumb-up", "dislike": "thumb-down"}

        self._session_id = id(self)

        if logs_filename is not None:
            self._logs = ChatLogs(filename=logs_filename)
            interface.post_hook = on_message

        llm = llm or self.llm
        instantiated = []
        self._analyses = []
        for agent in agents or self.agents:
            if not isinstance(agent, Agent):
                kwargs = {"llm": llm} if agent.llm is None else {}
                agent = agent(interface=interface, **kwargs)
            if agent.llm is None:
                agent.llm = llm
            # must use the same interface or else nothing shows
            agent.interface = interface
            if isinstance(agent, AnalysisAgent):
                self._analyses.extend(agent.analyses)
            instantiated.append(agent)

        super().__init__(llm=llm, agents=instantiated, interface=interface, logs_filename=logs_filename, **params)
        interface.send(
            "Welcome to LumenAI; get started by clicking a suggestion or type your own query below!",
            user="Help", respond=False,
        )
        interface.button_properties={
            "suggest": {"callback": self._create_suggestion, "icon": "wand"},
            "undo": {"callback": on_undo},
            "rerun": {"callback": on_rerun},
        }
        self._add_suggestions_to_footer(self.suggestions)

        self._current_agent = Markdown("## No agent active", margin=0)

        notebook_button = FileDownload(
            icon="notebook",
            button_type="success",
            callback=download_notebook,
            filename="Lumen_ai.ipynb",
            sizing_mode="stretch_width",
        )

        if "current_source" in memory and "available_sources" not in memory:
            memory["available_sources"] = {memory["current_source"]}
        elif "current_source" not in memory and "available_sources" in memory:
            memory["current_source"] = memory["available_sources"][0]
        elif "available_sources" not in memory:
            memory["available_sources"] = set([])

        self._controls = Column(
            notebook_button, *self.sidebar_widgets, self._current_agent, Tabs(("Memory", memory))
        )

    def _add_suggestions_to_footer(
        self,
        suggestions: list[str],
        num_objects: int = 1,
        inplace: bool = True,
        analysis: bool = False,
        append_demo: bool = True,
        hide_after_use: bool = True
    ):
        async def hide_suggestions(_=None):
            if len(self.interface.objects) > num_objects:
                suggestion_buttons.visible = False

        async def use_suggestion(event):
            button = event.obj
            with button.param.update(loading=True), self.interface.active_widget.param.update(loading=True):
                contents = button.name
                if hide_after_use:
                    suggestion_buttons.visible = False
                    if event.new > 1:  # prevent double clicks
                        return
                if analysis:
                    for agent in self.agents:
                        if isinstance(agent, AnalysisAgent):
                            break
                    else:
                        return
                    await agent.invoke([{'role': 'user', 'content': contents}], agents=self.agents)
                    self._add_analysis_suggestions()
                else:
                    self.interface.send(contents)

        async def run_demo(event):
            if hide_after_use:
                suggestion_buttons.visible = False
                if event.new > 1:  # prevent double clicks
                    return
            with self.interface.active_widget.param.update(loading=True):
                for demo_message in self.demo_inputs:
                    while self.interface.disabled:
                        await asyncio.sleep(1.25)
                    self.interface.active_widget.value = demo_message
                    await asyncio.sleep(2)

        suggestion_buttons = FlexBox(
            *[
                Button(
                    name=suggestion,
                    button_style="outline",
                    on_click=use_suggestion,
                    margin=5,
                )
                for suggestion in suggestions
            ],
            margin=(5, 5),
        )

        if append_demo and self.demo_inputs:
            suggestion_buttons.append(Button(
                name="Show a demo",
                button_type="primary",
                on_click=run_demo,
                margin=5,
            ))

        message = self.interface.objects[-1]
        if inplace:
            message.footer_objects = [suggestion_buttons]

        self.interface.param.watch(hide_suggestions, "objects")
        return message

    def _add_analysis_suggestions(self):
        pipeline = memory['current_pipeline']
        current_analysis = memory.get("current_analysis")
        allow_consecutive = getattr(current_analysis, '_consecutive_calls', True)
        applicable_analyses = []
        for analysis in self._analyses:
            if analysis.applies(pipeline) and (allow_consecutive or analysis is not type(current_analysis)):
                applicable_analyses.append(analysis)
        self._add_suggestions_to_footer(
            [f"Apply {analysis.__name__}" for analysis in applicable_analyses],
            append_demo=False,
            analysis=True,
            hide_after_use=False,
            num_objects=len(self.interface.objects),
        )

    async def _invalidate_memory(self, messages):
        table = memory.get("current_table")
        if not table:
            return

        source = memory.get("current_source")
        if table not in source:
            sources = [src for src in memory.get('available_sources', []) if table in src]
            if sources:
                memory['current_source'] = source = sources[0]
            else:
                raise KeyError(f'Table {table} could not be found in available sources.')

        try:
            spec = get_schema(source, table=table)
        except Exception:
            # If the selected table cannot be fetched we should invalidate it
            spec = None
        sql = memory.get("current_sql")
        system = render_template("check_validity.jinja2", table=table, spec=spec, sql=sql, analyses=self._analyses)
        with self.interface.add_step(title="Checking memory...", user="Assistant") as step:
            output = await self.llm.invoke(
                messages=messages,
                system=system,
                response_model=Validity,
            )
            step.stream(output.correct_assessment, replace=True)
            step.success_title = f"{output.is_invalid.title()} needs refresh" if output.is_invalid else "Memory still valid"

        if output and output.is_invalid:
            if output.is_invalid == "table":
                memory.pop("current_table", None)
                memory.pop("current_data", None)
                memory.pop("current_sql", None)
                memory.pop("current_pipeline", None)
                memory.pop("closest_tables", None)
                print("\033[91mInvalidated from memory.\033[0m")
            elif output.is_invalid == "sql":
                memory.pop("current_sql", None)
                memory.pop("current_data", None)
                memory.pop("current_pipeline", None)
                print("\033[91mInvalidated SQL from memory.\033[0m")

    async def _create_suggestion(self, instance, event):
        messages = self.interface.serialize(custom_serializer=self._serialize)[-3:-1]
        response = self.llm.stream(
            messages,
            system="Generate a follow-up question that a user might ask; ask from the user POV",
        )
        try:
            self.interface.disabled = True
            async for output in response:
                self.interface.active_widget.value_input = output
        finally:
            self.interface.disabled = False

    async def _chat_invoke(self, contents: list | str, user: str, instance: ChatInterface):
        print("\033[94mNEW\033[0m" + "-" * 100)
        await self.invoke(contents)

    @staticmethod
    def _create_agent_model(agent_names):
        agent_model = create_model(
            "RelevantAgent",
            chain_of_thought=(
                str,
                FieldInfo(
                    description="Explain in your own words, what the user wants."
                ),
            ),
            agent=(
                Literal[agent_names],
                FieldInfo(default=..., description="The most relevant agent to use.")
            ),
        )
        return agent_model

    @retry_llm_output()
    async def _create_valid_agent(self, messages, system, agent_model, errors=None):
        if errors:
            errors = '\n'.join(errors)
            messages += [{"role": "user", "content": f"\nExpertly resolve these issues:\n{errors}"}]

        out = await self.llm.invoke(
            messages=messages,
            system=system,
            response_model=agent_model
        )
        return out

    async def _choose_agent(self, messages: list | str, agents: list[Agent]):
        agents = [agent for agent in agents if agent.applies()]
        agent_names = tuple(sagent.name[:-5] for sagent in agents)
        if len(agent_names) == 0:
            raise ValueError("No agents available to choose from.")
        if len(agent_names) == 1:
            return agent_names[0]
        self._current_agent.object = "## **Current Agent**: [Lumen.ai](https://lumen.holoviz.org/)"
        agent_model = self._create_agent_model(agent_names)

        for agent in agents:
            if isinstance(agent, AnalysisAgent):
                analyses = "\n".join(
                    f"- `{analysis.__name__}`: {(analysis.__doc__ or '').strip()}"
                    for analysis in agent.analyses if analysis._callable_by_llm
                )
                agent.__doc__ = f"Available analyses include:\n{analyses}\nSelect this agent to perform one of these analyses."
                break

        system = render_template(
            "pick_agent.jinja2", agents=agents, current_agent=self._current_agent.object
        )
        return await self._create_valid_agent(messages, system, agent_model)

    async def _get_agent(self, messages: list | str):
        if len(self.agents) == 1:
            return self.agents[0]
        agent_types = tuple(agent.name[:-5] for agent in self.agents)
        agents = {agent.name[:-5]: agent for agent in self.agents}

        if len(agent_types) == 1:
            agent = agent_types[0]
        else:
            with self.interface.add_step(title="Selecting relevant agent...", user="Assistant") as step:
                output = await self._choose_agent(messages, self.agents)
                step.stream(output.chain_of_thought, replace=True)
                agent = output.agent
                step.success_title = f"Selected {agent}"

        if agent is None:
            return None

        print(
            f"Assistant decided on \033[95m{agent!r}\033[0m"
        )
        selected = subagent = agents[agent]
        agent_chain = []
        while unmet_dependencies := tuple(
            r for r in await subagent.requirements(messages) if r not in memory
        ):
            with self.interface.add_step(title="Solving dependency chain...") as step:
                step.stream(f"Found {len(unmet_dependencies)} unmet dependencies: {', '.join(unmet_dependencies)}")
                print(f"\033[91m### Unmet dependencies: {unmet_dependencies}\033[0m")
                subagents = [
                    agent
                    for agent in self.agents
                    if any(ur in agent.provides for ur in unmet_dependencies)
                ]
                output = await self._choose_agent(messages, subagents)
                subagent_name = output.agent if not isinstance(output, str) else output
                if subagent_name is None:
                    continue
                subagent = agents[subagent_name]
                agent_chain.append((subagent, unmet_dependencies))
                step.success_title = f"Solved a dependency with {subagent_name}"
        for subagent, deps in agent_chain[::-1]:
            with self.interface.add_step(title="Choosing subagent...") as step:
                step.stream(f"Assistant decided the {subagent.name[:-5]!r} will provide {', '.join(deps)}.")
                self._current_agent.object = f"## **Current Agent**: {subagent.name[:-5]}"

                if isinstance(subagent, SQLAgent):
                    custom_messages = messages.copy()
                    custom_agent = next((agent for agent in self.agents if isinstance(agent, AnalysisAgent)), None)
                    if custom_agent:
                        custom_analysis_doc = custom_agent.__doc__.replace("Available analyses include:\n", "")
                        custom_message = (
                            f"Avoid doing the same analysis as any of these custom analyses: {custom_analysis_doc} "
                            f"Most likely, you'll just need to do a simple SELECT * FROM {{table}};"
                        )
                        custom_messages.append({"role": "user", "content": custom_message})
                    await subagent.answer(custom_messages)
                else:
                    await subagent.answer(messages)
                step.success_title = f"Selected {subagent.name[:-5]}"
        return selected

    def _serialize(self, obj):
        if isinstance(obj, (Tabs, Column)):
            for o in obj:
                if isinstance(obj, ChatStep) and not obj.title.startswith("Selected"):
                    # only want the chain of thoughts from the selected agent
                    continue
                if hasattr(o, "visible") and o.visible:
                    break
            else:
                return ""
            return self._serialize(o)

        if isinstance(obj, HTML) and 'catalog' in obj.tags:
            return f"Summarized table listing: {obj.object[:30]}"

        if hasattr(obj, "object"):
            obj = obj.object
        elif hasattr(obj, "value"):
            obj = obj.value
        return str(obj)

    async def invoke(self, messages: list | str) -> str:
        messages = self.interface.serialize(custom_serializer=self._serialize)[-4:]
        await self._invalidate_memory(messages[-2:])
        agent = await self._get_agent(messages[-3:])
        if agent is None:
            msg = (
                "Assistant could not settle on an agent to perform the requested query. "
                "Please restate your request."
            )
            self.interface.stream(msg, user='Lumen')
            return msg

        self._current_agent.object = f"## **Current Agent**: {agent.name[:-5]}"
        print("\n\033[95mMESSAGES:\033[0m")
        for message in messages:
            print(f"{message['role']!r}: {message['content']}")
            print("ENTRY" + "-" * 10)

        print("\n\033[95mAGENT:\033[0m", agent, messages[-3:])

        kwargs = {}
        if isinstance(agent, AnalysisAgent):
            kwargs["agents"] = self.agents
        await agent.invoke(messages[-3:], **kwargs)
        self._current_agent.object = "## No agent active"
        if "current_pipeline" in agent.provides:
            self._add_analysis_suggestions()
        print("\033[92mDONE\033[0m", "\n\n")

    def controls(self):
        return self._controls

    def __panel__(self):
        return self.interface



import io
import tempfile

import panel as pn
import param  # type: ignore
import yaml

from ..dashboard import Dashboard
from ..state import state as lm_state
from ..util import is_ref, resolve_module_reference
from .base import WizardItem
from .gallery import Gallery, GalleryItem
from .state import state


class LauncherGalleryItem(GalleryItem):
    """
    Gallery Item corresponding to a specific Launcher component.
    """

    icon = param.String(doc="The FontAwesome icon to render.")

    launcher = param.Parameter(precedence=-1, doc="The Launcher instance.")

    selected = param.Boolean(default=False, doc="Whether the item was selected.")

    spec = param.Dict(doc="The specification for the Launcher instance.")

    _template = """
    <div id="launcher-item" onclick="${_select}" style="display: flex; flex-direction: column; height: 100%; justify-content: space-between;">
      <div style="font-size: 1.25em; font-weight: bold;">{{ name }}</div>
      <div style="text-align: center; font-size: 8em;">
        <i id="icon" class="fas {{ launcher.icon }}"></i>
      </div>
      <p style="height: 4em;">${description}</p>
    </div>
    """

    def __init__(self, launcher, **params):
        params['name'] = launcher.name.replace('Launcher', '')
        description = ''
        for line in launcher.__doc__.split('\n'):
            if line.strip() and not line.startswith('params('):
                description = line
                break
        super().__init__(launcher=launcher, description=description, **params)

    def _select(self, event):
        self.selected = True


class Launcher(WizardItem):
    """
    A WizardItem that can be used to export or launch the dashboard
    once it has been built.
    """

    __abstract = True


class LauncherGallery(WizardItem, Gallery):
    "Launch or download your dashboard specification."

    auto_advance = param.Boolean(default=True)

    path = param.Foldername()

    spec = param.Dict(default={}, precedence=-1)

    _template = """
    <span style="font-size: 2em;">{{ __doc__ }}</span>
    <fast-divider style="margin: 1em 0;"></fast-divider>
    <div id="items" style="margin: 1em 0; display: flex; flex-wrap: wrap; gap: 1em;">
    {% for item in items.values() %}
      <fast-card id="launcher-container" class="gallery-item" style="height: 290px; width: 350px; padding: 1em; margin: 0.2em;">
        ${item}
      </fast-card>
    {% endfor %}
    </div>
    """

    _gallery_item = LauncherGalleryItem

    def __init__(self, **params):
        self.builder = params.pop('builder')
        super().__init__(**params)
        if not self.items:
            self.items['local'] = LauncherGalleryItem(launcher=LocalLauncher)
            self.items['yaml'] = LauncherGalleryItem(launcher=YAMLLauncher)
        for item in self.items.values():
            item.param.watch(self._advance, 'selected')

    def _preprocess_kwargs(self, kwargs):
        spec = kwargs.get('spec', {})
        launcher_type = spec.pop('type')
        launchers = param.concrete_descendents(Launcher)
        if launcher_type in launchers:
            cls = launchers[launcher_type]
        else:
            cls = resolve_module_reference(launcher_type, Launcher)
        for key, value in list(spec.items()):
            if is_ref(value):
                spec[key] = lm_state.resolve_reference(value)
        kwargs['launcher'] = cls
        return kwargs

    def _advance(self, event):
        launcher = event.obj.launcher(**event.obj.spec)
        items = self.builder.wizard.items
        if isinstance(items[-1], Launcher):
            items = items[:-1]
        items.append(launcher)
        self.builder.wizard.items = items
        self.builder.wizard.next_disable = False
        self.builder.wizard._next()


class LocalLauncher(Launcher):
    """
    Launches the dashboard locally.
    """

    icon = param.String(default='fa-window-maximize')

    _template = """
    <span style="font-size: 1.5em">Launcher</span>
    <p>{{ __doc__ }}</p>
    <fast-divider></fast-divider>
    <div style="width: 100%">
      <fast-button id="launch-button" style="margin: 2em auto;" onclick="${_launch}">Launch</fast-button>
    </div>
    """

    def _launch(self, event):
        with tempfile.NamedTemporaryFile('w', suffix='.yaml', delete=False) as f:
            f.file.write(yaml.dump(state.spec))
        dashboard = Dashboard(f.name)
        dashboard._render_dashboard()
        dashboard.show()


class YAMLLauncher(Launcher):
    """
    Download or copy the dashboard yaml specification.
    """

    download = param.ClassSelector(class_=pn.widgets.FileDownload)

    editor = param.ClassSelector(class_=pn.pane.JSON)

    icon = param.String(default='fa-file-code')

    _template = """
    <span style="font-size: 1.5em">Launcher</span>
    <p>{{ __doc__ }}</p>
    <fast-divider></fast-divider>
    <div id="download">${download}</div>
    <div id="yaml">${editor}</div>
    """

    def __init__(self, **params):
        params['download'] = pn.widgets.FileDownload(
            name='Download Dashboard Yaml',
            callback=self._download_data,
            filename='dashboard.yaml',
            margin=(5, 0)
        )
        params['editor'] = pn.pane.JSON(
            sizing_mode='stretch_both',
            margin=0,
            depth=-1
        )
        super().__init__(**params)

    def _download_data(self):
        sio = io.StringIO()
        text = yaml.dump(state.spec)
        sio.write(text)
        sio.seek(0)
        return sio

    @param.depends('active', watch=True)
    def _update_spec(self):
        self.spec = state.spec
        self.editor.object = state.spec



import datetime
import inspect
import warnings

from typing import (
    Any, Callable, Literal, Optional, TypeVar, Union,
)

import param

from pydantic import BaseConfig, BaseModel, create_model
from pydantic.color import Color
from pydantic.fields import FieldInfo, PydanticUndefined

DATE_TYPE = Union[datetime.datetime, datetime.date]
PARAM_TYPE_MAPPING: dict[param.Parameter, type] = {
    param.String: str,
    param.Integer: int,
    param.Number: float,
    param.Boolean: bool,
    param.Event: bool,
    param.Date: DATE_TYPE,
    param.DateRange: tuple[DATE_TYPE],
    param.CalendarDate: DATE_TYPE,
    param.CalendarDateRange: tuple[DATE_TYPE],
    param.Parameter: object,
    param.Color: Color,
    param.Callable: Callable,
}
PandasDataFrame = TypeVar("PandasDataFrame")


class ArbitraryTypesModel(BaseModel):
    """
    A Pydantic model that allows arbitrary types.
    """

    class Config(BaseConfig):
        arbitrary_types_allowed = True


def _create_literal(obj: list[Union[str, type]]) -> type:
    """
    Create a literal type from a list of objects.
    """
    enum = []
    for item in obj:
        if item is None:
            continue
        elif isinstance(item, str):
            enum.append(item)
        else:
            enum.append(item.__name__)
    if enum:
        return Literal[tuple(enum)]
    else:
        return str


def _get_model(type_, created_models: dict[str, BaseModel]) -> Any:
    try:
        if issubclass(type_, param.Parameterized):
            type_ = created_models.get(type_.__name__, type_.__name__)
    except TypeError:
        pass
    return type_


def parameter_to_field(
    parameter: param.Parameter, created_models: dict[str, BaseModel],
    literals: list[str] | None
) -> (type, FieldInfo):
    """
    Translate a parameter to a pydantic field.
    """
    param_type = parameter.__class__
    description = " ".join(parameter.doc.split()) if parameter.doc else None
    field_info = FieldInfo(description=description)
    if not literals and hasattr(parameter, 'get_range'):
        literals = list(parameter.get_range())

    if param_type in PARAM_TYPE_MAPPING:
        type_ = PARAM_TYPE_MAPPING[param_type]
        field_info.default = parameter.default
    elif param_type is param.ClassSelector:
        type_ = _get_model(parameter.class_, created_models)
        if isinstance(type_, tuple):
            type_ = Union[tuple([PARAM_TYPE_MAPPING.get(t, t) for t in type_])]
        if parameter.default is not None:
            default_factory = parameter.default
            if not callable(default_factory):
                default_factory = type(default_factory)
            field_info.default_factory = default_factory
    elif param_type in [param.List, param.ListSelector]:
        type_ = list
        if parameter.default is not None:
            field_info.default_factory = parameter.default
        if param_type is param.List and parameter.item_type:
            type_ = list[_get_model(parameter.item_type, created_models)]
        elif param_type is param.ListSelector:
            type_ = list[_create_literal(literals)]
    elif param_type is param.Dict:
        type_ = dict
        if parameter.default == {}:
            field_info.default_factory = dict
        elif parameter.default is not None:
            field_info.default_factory = parameter.default
    elif param_type in [param.Selector, param.ObjectSelector]:
        field_info.default = parameter.default
        if literals:
            type_ = _create_literal(literals)
        elif parameter.objects:
            type_ = _create_literal(parameter.objects)
        else:
            type_ = str
    elif issubclass(param_type, param.DataFrame):
        type_ = PandasDataFrame
    elif parameter.name == "align":
        type_ = _create_literal(["auto", "start", "center", "end"])
    elif parameter.name == "aspect_ratio":
        type_ = Union[Literal["auto"], float]
    elif parameter.name == "margin":
        type_ = Union[float, tuple[float, float], tuple[float, float, float, float]]
    else:
        raise NotImplementedError(
            f"Parameter {parameter.name!r} of {param_type.__name__!r} not supported"
        )

    if hasattr(parameter, "bounds") and parameter.bounds and type_ in [int, float]:
        try:
            field_info.ge = parameter.bounds[0]
            field_info.le = parameter.bounds[1]
        except Exception:
            pass

    if parameter.allow_None:
        type_ = Optional[type_]

    return type_, field_info


def param_to_pydantic(
    parameterized: type[param.Parameterized],
    base_model: type[BaseModel] = ArbitraryTypesModel,
    created_models: Optional[dict[str, BaseModel]] = None,
    schema: Optional[dict[str, Any]] = None,
    excluded: Union[str, list[str]] = "_internal_params",
    extra_fields: Optional[dict[str, tuple[type, FieldInfo]]] = None,
) -> dict[str, BaseModel]:
    """
    Translate a param Parameterized to a Pydantic BaseModel.

    Parameters
    ----------
    parameterized : Type[param.Parameterized]
        The parameterized class to translate.
    base_model : Type[BaseModel], optional
        The base model to use, by default ArbitraryTypesModel
    created_models : Optional[Dict[str, BaseModel]], optional
        A dictionary of already created models, by default None
    excluded : Union[str, List[str]], optional
        A list of parameters to exclude. If a string is provided,
        it looks up the parameter on the parameterized class and
        excludes the parameter's value, by default "_internal_params".
    extra_fields : Optional[Dict[str, Tuple[Type, FieldInfo]]], optional
        Extra fields to add to the model, by default None
    """
    parameterized_name = parameterized.__name__
    if created_models is None:
        created_models = {}
    if parameterized_name in created_models:
        return created_models

    if isinstance(excluded, str) and hasattr(parameterized, excluded):
        excluded = getattr(parameterized, excluded)

    parameterized_signature = inspect.signature(parameterized.__init__)
    required_args = [
        arg.name
        for arg in parameterized_signature.parameters.values()
        if arg.name not in ["self", "params"] and arg.default == inspect._empty
    ]
    field_params = list(getattr(parameterized, '_field_params', []))

    fields = {}
    for parameter_name in parameterized.param:
        if parameter_name in excluded or parameterized_name.lower() == parameter_name.lower():
            continue
        parameter = parameterized.param[parameter_name]
        if hasattr(parameter, "class_") and hasattr(parameter.class_, "param"):
            parameter_class_name = parameter.class_.__name__
            if parameterized_name != parameter_class_name:
                param_to_pydantic(parameter.class_, created_models=created_models)

        literals = list(schema) if schema and parameter_name in field_params else None

        type_, field_info = parameter_to_field(parameter, created_models, literals)
        if parameter_name == "schema":
            field_info.alias = "schema"
            parameter_name = "schema_"
        elif parameter_name == "copy":
            field_info.alias = "copy"
            parameter_name = "copy_"
        elif parameter_name[0] == "_":
            field_info.alias = parameter_name
            parameter_name = parameter_name.lstrip("_")

        if parameter_name in required_args:
            field_info.default = PydanticUndefined
        elif field_info.default == PydanticUndefined and not field_info.default_factory:
            field_info.default = None
        fields[parameter_name] = (type_, field_info)

    fields["__parameterized__"] = (type, parameterized)

    if extra_fields:
        fields.update(extra_fields)

    # ignore RuntimeWarning: fields may not start with an underscore
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", RuntimeWarning)
        pydantic_model = create_model(
            parameterized.__name__, __base__=base_model, **fields
        )
    created_models[pydantic_model.__name__] = pydantic_model
    return created_models


def pydantic_to_param(model: BaseModel) -> param.Parameterized:
    """
    Serialize a Pydantic model to a Parameterized instance.

    Parameters
    ----------
    model : BaseModel
        The model to serialize.
    """
    kwargs = {}
    for key, value in dict(model).items():
        if isinstance(value, BaseModel):
            kwargs[key] = pydantic_to_param(value)
        elif isinstance(value, Color):
            kwargs[key] = value.as_hex()
        elif isinstance(value, dict):
            sub_kwargs = {}
            for k, v in value.items():
                if isinstance(v, BaseModel):
                    sub_kwargs[k] = pydantic_to_param(v)
                else:
                    sub_kwargs[k] = v
            kwargs[key] = sub_kwargs
        elif key == "table":
            # couldn't convince AI to provide path to file for `tables`
            # but remove the extension for `table`
            kwargs[key] = value.split(".")[0]
        else:
            kwargs[key] = value

    parameterized = model.__parameterized__(**kwargs)
    return parameterized



import datetime as dt
import os

import pandas as pd
import pytest

from lumen.transforms.sql import SQLGroupBy

try:
    from lumen.sources.duckdb import DuckDBSource
    pytestmark = pytest.mark.xdist_group("duckdb")
except ImportError:
    pytestmark = pytest.mark.skip(reason="Duckdb is not installed")



@pytest.fixture
def duckdb_source():
    root = os.path.dirname(__file__)
    duckdb_source = DuckDBSource(
        initializers=[
            "INSTALL sqlite;",
            "LOAD sqlite;",
            f"SET home_directory='{root}';"
        ],
        root=root,
        sql_expr="SELECT A, B, C, D::TIMESTAMP_NS AS D FROM {table}",
        tables={
            'test_sql': f"sqlite_scan('{root + '/test.db'}', 'mixed')",
            'test_sql_with_none': f"sqlite_scan('{root + '/test.db'}', 'mixed_none')",
        }
    )
    return duckdb_source


@pytest.fixture
def duckdb_memory_source(mixed_df):
    source = DuckDBSource(uri=':memory:', ephemeral=True)
    source._connection.from_df(mixed_df).to_view('mixed')
    return source


def test_duckdb_resolve_module_type():
    assert DuckDBSource._get_type('lumen.sources.duckdb.DuckDBSource') is DuckDBSource
    assert DuckDBSource.source_type == 'duckdb'


def test_duckdb_get_tables(duckdb_source, source_tables):
    tables = duckdb_source.get_tables()
    assert not len(set(tables) - set(source_tables.keys()))
    for table in tables:
        pd.testing.assert_frame_equal(
            duckdb_source.get(table),
            source_tables[table],
        )


def test_duckdb_get_schema(duckdb_source):
    expected_sql = {
        'A': {'inclusiveMaximum': 4.0, 'inclusiveMinimum': 0.0, 'type': 'number'},
        'B': {'inclusiveMaximum': 1.0, 'inclusiveMinimum': 0.0, 'type': 'number'},
        'C': {'enum': ['foo1', 'foo2', 'foo3', 'foo4', 'foo5'], 'type': 'string'},
        'D': {
            'format': 'datetime',
            'inclusiveMaximum': '2009-01-07 00:00:00',
            'inclusiveMinimum': '2009-01-01 00:00:00',
            'type': 'string'
        }
    }
    source = duckdb_source.get_schema('test_sql')
    source["C"]["enum"].sort()
    assert source == expected_sql
    assert list(duckdb_source._schema_cache.keys()) == ['test_sql']


def test_duckdb_get_schema_with_none(duckdb_source):
    enum = ['foo1', None, 'foo3', 'foo5']
    expected_sql = {
        'A': {'inclusiveMaximum': 4.0, 'inclusiveMinimum': 0.0, 'type': 'number'},
        'B': {'inclusiveMaximum': 1.0, 'inclusiveMinimum': 0.0, 'type': 'number'},
        'C': {'enum': enum, 'type': 'string'},
        'D': {
            'format': 'datetime',
            'inclusiveMaximum': '2009-01-07 00:00:00',
            'inclusiveMinimum': '2009-01-01 00:00:00',
            'type': 'string'
        }
    }
    source = duckdb_source.get_schema('test_sql_with_none')
    source["C"]["enum"].sort(key=enum.index)
    assert source == expected_sql
    assert list(duckdb_source._schema_cache.keys()) == ['test_sql_with_none']


def test_duckdb_get_schema_cache(duckdb_source):
    duckdb_source.get_schema('test_sql')
    assert 'test_sql' in duckdb_source._schema_cache


@pytest.mark.parametrize(
    "table_column_value_type", [
        ('test_sql', 'A', 1, 'single_value'),
        ('test_sql', 'A', (1, 3), 'range'),
        ('test_sql', 'A', [(0, 1), (3, 4)], 'range_list'),
        ('test_sql', 'C', 'foo2', 'single_value'),
        ('test_sql', 'C', ['foo1', 'foo3'], 'list'),
        ('test_sql', 'D', dt.datetime(2009, 1, 2), 'single_value'),
        ('test_sql', 'D', (dt.datetime(2009, 1, 2), dt.datetime(2009, 1, 5)), 'range'),
        ('test_sql', 'D', dt.date(2009, 1, 2), 'date'),
        ('test_sql', 'D', (dt.date(2009, 1, 2), dt.date(2009, 1, 5)), 'date_range'),
        ('test_sql_with_none', 'C', None, 'single_value'),
        ('test_sql_with_none', 'C', [None, 'foo5'], 'list'),
    ]
)
@pytest.mark.parametrize("dask", [True, False])
def test_duckdb_filter(duckdb_source, table_column_value_type, dask, expected_filtered_df):
    table, column, value, _ = table_column_value_type
    kwargs = {column: value}
    filtered = duckdb_source.get(table, __dask=dask, **kwargs)
    pd.testing.assert_frame_equal(filtered, expected_filtered_df.reset_index(drop=True))


@pytest.mark.flaky(reruns=3)
def test_duckdb_transforms(duckdb_source, source_tables):
    df_test_sql = source_tables['test_sql']
    transforms = [SQLGroupBy(by=['B'], aggregates={'SUM': 'A'})]
    transformed = duckdb_source.get('test_sql', sql_transforms=transforms)
    expected = df_test_sql.groupby('B')['A'].sum().reset_index()
    pd.testing.assert_frame_equal(transformed, expected)


@pytest.mark.flaky(reruns=3)
def test_duckdb_transforms_cache(duckdb_source, source_tables):
    df_test_sql = source_tables['test_sql']
    transforms = [SQLGroupBy(by=['B'], aggregates={'SUM': 'A'})]

    duckdb_source.get('test_sql', sql_transforms=transforms)
    cache_key = duckdb_source._get_key('test_sql', sql_transforms=transforms)
    assert cache_key in duckdb_source._cache

    expected = df_test_sql.groupby('B')['A'].sum().reset_index()
    pd.testing.assert_frame_equal(duckdb_source._cache[cache_key], expected)

    cache_key = duckdb_source._get_key('test_sql', sql_transforms=transforms)
    assert cache_key in duckdb_source._cache


def test_duckdb_clear_cache(duckdb_source):
    duckdb_source.get('test_sql')
    duckdb_source.get_schema('test_sql')
    assert len(duckdb_source._cache) == 1
    assert len(duckdb_source._schema_cache) == 1
    duckdb_source.clear_cache()
    assert len(duckdb_source._cache) == 0
    assert len(duckdb_source._schema_cache) == 0


def test_duckdb_source_ephemeral_roundtrips(duckdb_memory_source, mixed_df):
    source = DuckDBSource.from_spec(duckdb_memory_source.to_spec())
    df = source.get('mixed')
    for col in df.columns:
        assert (df[col]==mixed_df[col]).all()


def test_duckdb_source_mirrors_source(duckdb_source):
    mirrored = DuckDBSource(uri=':memory:', mirrors={'mixed': (duckdb_source, 'test_sql')})
    pd.testing.assert_frame_equal(duckdb_source.get('test_sql'), mirrored.get('mixed'))



from pathlib import Path

import panel as pn


class LlmSetupError(Exception):
    """
    Raised when an error occurs during the setup of the LLM.
    """


THIS_DIR = Path(__file__).parent

FUZZY_TABLE_LENGTH = 10

GETTING_STARTED_SUGGESTIONS = [
    "What datasets do you have?",
    "Tell me about the dataset.",
    "Create a plot of the dataset.",
    "Find the min and max of the values.",
]

DEMO_MESSAGES = [
    "What data is available?",
    "Can I see the first one?",
    "Tell me about the dataset.",
    "What could be interesting to analyze?",
    "Perform a SQL query on one of these.",
    "Show it to me as a scatter plot.",
]

DEFAULT_EMBEDDINGS_PATH = Path("embeddings")

UNRECOVERABLE_ERRORS = (
    ImportError,
    LlmSetupError,
    RecursionError,
)

pn.chat.ChatStep.min_width = 350



from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar
from weakref import WeakKeyDictionary

import panel as pn

from panel import state
from panel.viewable import Viewer

from ..base import Component
from ..config import SessionCache

if TYPE_CHECKING:
    from bokeh.document import Document


class _Memory(SessionCache, Viewer):

    _views: ClassVar[WeakKeyDictionary[Document, Any]] = WeakKeyDictionary()

    _global_view = None

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        if state.curdoc in self._views or (state.curdoc is None and self._global_view is not None):
            self._update_view(key, value)

    def _render_item(self, key, item):
        if isinstance(item, Component):
            item = item.to_spec()
            if 'password' in item:
                item['password'] = '•'*len(item['password'])
        if isinstance(item, str):
            item = f'```yaml\n{item}\n```'
        return pn.panel(item, name=key, sizing_mode='stretch_width', styles={'overflow': 'scroll'})

    def _render_memories(self):
        return pn.Accordion(*(
            self._render_item(name, item)
            for name, item in self._curcontext.items()
        ), sizing_mode='stretch_width', active=list(range(len(self._curcontext))))

    def _update_view(self, key, value):
        view = self._views[state.curdoc] if state.curdoc else self._global_view
        accordion = view[0]
        new_item = self._render_item(key, value)
        i = 0
        for i, item in enumerate(accordion):
            if item.name == key:
                accordion[i] = new_item
                break
        else:
            accordion.append(new_item)
            accordion.active = accordion.active + [i+1]

    def __panel__(self):
        view = pn.Column(self._render_memories())
        if state.curdoc is None:
            self._global_view = view
        else:
            self._views[state.curdoc] = view
        return view


memory = _Memory()



import panel as pn
import param

from ..base import Component
from .controls import SourceControls
from .memory import memory
from .utils import get_schema


class Analysis(param.ParameterizedFunction):
    """
    The purpose of an Analysis is to perform custom actions tailored
    to a particular use case and/or domain. It is a ParameterizedFunction
    which implements an applies method to confirm whether the analysis
    applies to a particular dataset and a __call__ method which is given
    a pipeline and must return another valid component.
    """

    agents = param.List(default=[], doc="List of available agents that can be invoked by the analysis.")

    autorun = param.Boolean(default=True, doc="Whether to automatically run the analysis.")

    columns = param.List(default=[], constant=True, doc="""
       The columns required for the analysis. May use tuples to declare that one of
       the columns must be present.""")

    _run_button = param.Parameter(default=None)

     # Whether to allow the analysis to be called multiple times in a row
    _consecutive_calls = False

    _callable_by_llm = True

    _field_params = []

    @classmethod
    def applies(cls, pipeline) -> bool:
        applies = True
        for col in cls.columns:
            if isinstance(col, tuple):
                applies &= any(c in pipeline.data.columns for c in col)
            else:
                applies &= col in pipeline.data.columns
        return applies

    def controls(self):
        config_options = [p for p in self.param if p not in Analysis.param]
        if config_options:
            return pn.Param(self.param, parameters=config_options)

    def __call__(self, pipeline) -> Component:
        return pipeline


class Join(Analysis):

    autorun = param.Boolean(default=False, doc="Whether to automatically run the analysis.")

    table_name = param.String(doc="The name of the table to join.")

    index_col = param.String(doc="The column to join on in the left table.")

    context = param.String(doc="Additional context to provide to the LLM.")

    _callable_by_llm = False

    _previous_table = param.Parameter()

    _previous_source = param.Parameter()

    def _update_table_name(self, event):
        self.table_name = event.new

    def controls(self):
        self._source_controls = SourceControls()
        self._run_button = self._source_controls._add_button
        self._source_controls.param.watch(self._update_table_name, "_last_table")

        source = memory.get("current_source")
        table = memory.get("current_table")
        self._previous_source = source
        self._previous_table = table
        columns = list(get_schema(source, table=table).keys())
        index_col = pn.widgets.AutocompleteInput.from_param(
            self.param.index_col, options=columns, name="Join on",
            placeholder="Start typing column name", search_strategy="includes",
            case_sensitive=False, restrict=False
        )
        context = pn.widgets.TextInput.from_param(self.param.context, name="Context")
        controls = pn.FlexBox(
            index_col,
            context,
            self._source_controls,
        )
        return controls

    async def __call__(self, pipeline) -> Component:
        if self.table_name:
            agent = next(agent for agent in self.agents if type(agent).__name__ == "SQLAgent")
            content = f"Join these tables: '//{self._previous_source}//{self._previous_table}' and '//{memory['current_source']}//{self.table_name}'"
            if self.index_col:
                content += f" left join on {self.index_col}"
            else:
                content += " based on the closest matching columns"
            if self.context:
                content += f"\nadditional context:\n{self.context!r}"
            await agent.answer(messages=[{"role": "user", "content": content}])
            pipeline = memory["current_pipeline"]
        return pipeline



import asyncio

import panel as pn
import param
import yaml

from panel.viewable import Viewer
from param.parameterized import discard_events

from ..base import Component
from ..dashboard import load_yaml
from ..downloads import Download
from ..pipeline import Pipeline
from ..transforms.sql import SQLLimit
from ..views.base import Table


class LumenOutput(Viewer):

    active = param.Integer(default=1)

    component = param.ClassSelector(class_=Component)

    spec = param.String(allow_None=True)

    language = "yaml"

    def __init__(self, **params):
        if 'spec' not in params and 'component' in params and params['component'] is not None:
            component_spec = params['component'].to_spec()
            params['spec'] = yaml.safe_dump(component_spec)
        super().__init__(**params)
        code_editor = pn.widgets.CodeEditor(
            value=self.param.spec, language=self.language, sizing_mode="stretch_both",
        )
        code_editor.link(self, bidirectional=True, value='spec')
        copy_icon = pn.widgets.ButtonIcon(
            icon="copy", active_icon="check", toggle_duration=1000
        )
        copy_icon.js_on_click(
            args={"code_editor": code_editor},
            code="navigator.clipboard.writeText(code_editor.code);",
        )
        download_icon = pn.widgets.ButtonIcon(
            icon="download", active_icon="check", toggle_duration=1000
        )
        download_icon.js_on_click(
            args={"code_editor": code_editor},
            code="""
            var text = code_editor.code;
            var blob = new Blob([text], {type: 'text/plain'});
            var url = window.URL.createObjectURL(blob);
            var a = document.createElement('a');
            a.href = url;
            a.download = 'lumen_spec.yaml';
            document.body.appendChild(a); // we need to append the element to the dom -> otherwise it will not work in firefox
            a.click();
            a.parentNode.removeChild(a);  //afterwards we remove the element again
            """,
        )
        icons = pn.Row(copy_icon, download_icon)
        code_col = pn.Column(code_editor, icons, sizing_mode="stretch_both")
        placeholder = pn.Column(sizing_mode="stretch_width")
        self._tabs = pn.Tabs(
            ("Code", code_col),
            ("Output", placeholder),
            styles={'min-width': '100%', 'height': 'fit-content', 'min-height': '300px'},
            active=self.active,
            dynamic=True
        )
        self._tabs.link(self, bidirectional=True, active='active')
        self._rendered = False
        placeholder.objects = [
            pn.pane.ParamMethod(self._render_component, inplace=True)
        ]
        self._last_output = {}

    def _render_pipeline(self, pipeline):
        table = Table(
            pipeline=pipeline, pagination='remote',
            min_height=500, sizing_mode="stretch_both", stylesheets=[
                """
                .tabulator-footer {
                display: flex;
                text-align: left;
                padding: 0px;
                }
                """
            ]
        )
        download = Download(
            view=table, hide=False, filename=f'{pipeline.table}',
            format='csv'
        )
        download_pane = download.__panel__()
        download_pane.sizing_mode = 'fixed'
        controls = pn.Row(
            download_pane,
            styles={'position': 'absolute', 'right': '40px', 'top': '-35px'}
        )
        for sql_limit in pipeline.sql_transforms:
            if isinstance(sql_limit, SQLLimit):
                break
        else:
            sql_limit = None
        if sql_limit:
            limited = len(pipeline.data) == sql_limit.limit
            if limited:
                def unlimit(e):
                    sql_limit.limit = None if e.new else 1_000_000
                full_data = pn.widgets.Checkbox(
                    name='Full data', width=100, visible=limited
                )
                full_data.param.watch(unlimit, 'value')
                controls.insert(0, full_data)
        return pn.Column(controls, table)

    @param.depends('spec', 'active')
    async def _render_component(self):
        yield pn.indicators.LoadingSpinner(
            value=True, name="Rendering component...", height=50, width=50
        )

        if (self.active != (len(self._tabs)-1)) or self.spec is None:
            return

        if self.spec in self._last_output:
            yield self._last_output[self.spec]
            return
        elif self.component is None:
            yield pn.pane.Alert(
                "No component to render. Please complete the Config tab.",
                alert_type="warning",
            )
            return

        try:
            if self._rendered:
                yaml_spec = load_yaml(self.spec)
                self.component = type(self.component).from_spec(yaml_spec)
            if isinstance(self.component, Pipeline):
                output = self._render_pipeline(self.component)
            else:
                output = self.component.__panel__()
            self._rendered = True
            self._last_output.clear()
            self._last_output[self.spec] = output
            yield output
        except Exception as e:
            import traceback
            traceback.print_exc()
            yield pn.pane.Alert(
                f"```\n{e}\n```\nPlease press undo, edit the YAML, or continue chatting.",
                alert_type="danger",
            )

    def __panel__(self):
        return self._tabs

    def __repr__(self):
        return self.spec

    def __str__(self):
        return self.spec


class AnalysisOutput(LumenOutput):

    analysis = param.Parameter()

    pipeline = param.Parameter()

    def __init__(self, **params):
        if not params['analysis'].autorun:
            params['active'] = 0
        super().__init__(**params)
        controls = self.analysis.controls()
        if controls is not None:
            if self.analysis._run_button:
                run_button = self.analysis._run_button
                run_button.param.watch(self._rerun, 'clicks')
                self._tabs.insert(1, ('Config', controls))
            else:
                run_button = pn.widgets.Button(
                    icon='player-play', name='Run', on_click=self._rerun,
                    button_type='success', margin=(10, 0, 0 , 10)
                )
                self._tabs.insert(1, ('Config', pn.Column(controls, run_button)))
            with discard_events(self):
                self._tabs.active = 2 if self.analysis.autorun else 1
        self._rendered = True

    async def _rerun(self, event):
        with self._tabs.param.update(loading=True):
            if asyncio.iscoroutinefunction(self.analysis.__call__):
                view = await self.analysis(self.pipeline)
            else:
                view = await asyncio.to_thread(self.analysis, self.pipeline)
            self.component = view
            self._rendered = False
            spec = view.to_spec()
            self.param.update(
                spec=yaml.safe_dump(spec),
                active=2
            )


class SQLOutput(LumenOutput):

    language = "sql"

    @param.depends('spec', 'active')
    async def _render_component(self):
        yield pn.indicators.LoadingSpinner(
            value=True, name="Executing SQL query...", height=50, width=50
        )
        if self.active != 1:
            return

        pipeline = self.component
        if self.spec in self._last_output:
            yield self._last_output[self.spec]
            return

        try:
            if self._rendered:
                pipeline.source = pipeline.source.create_sql_expr_source(tables={pipeline.table: self.spec})
            output = self._render_pipeline(pipeline)
            self._rendered = True
            self._last_output.clear()
            self._last_output[self.spec] = output
            yield output
        except Exception as e:
            import traceback
            traceback.print_exc()
            yield pn.pane.Alert(
                f"```\n{e}\n```\nPlease press undo, edit the YAML, or continue chatting.",
                alert_type="danger",
            )

    def __panel__(self):
        return self._tabs
